# Author: AIMPED
# Date: 2023-March-12
# Description: This file contains the pipeline for de-identification of clinical notes


def maskText(merged, text):
    """
    It masks the actual chunks in the text with their entity labels.
    parameters:
    ----------------
    merged: list of dict
    text: str
    return:
    ----------------
    masked_text: str
    """
    text_mask = text[:]
    for i in range(len(merged) - 1, -1, -1):
        text_mask = text_mask[:merged[i]['begin']] + f"<<{merged[i]['entity']}>>" + text_mask[merged[i]['end']:]
    return text_mask


def fakedChunk(fake_csv_path, merged ):
    """ 
    Randomly select entities from fake.csv file and add them to entities list align with true labels.
    parameters:
    ----------------
    fake_csv_path: str
    merged: list of dict
    return:
    ----------------
    merged: list of dict
    """
    import pandas as pd
    import random

    fake_df = pd.read_csv(fake_csv_path, sep=',', encoding='utf8')
    for item in merged:
        item["faked_chunk"] = str(random.choice(fake_df[item['entity']]))
    return merged


def fakedText(merged, text):
    """
    Rewrite the text with faked chunks that comes from fakedChunk function.
    parameters:
    ----------------
    merged: list of dict
    text: str
    return:
    ----------------
    faked_text: str
    """
    
    faked_text = text[:]
    for i in range(len(merged) - 1, -1, -1):
        faked_text = faked_text[:merged[i]['begin']] + f"{merged[i]['faked_chunk']}" + faked_text[merged[i]['end']:]
    return faked_text


def deidentification(faked, masked, merged_results, text, fake_csv_path):
    """
    It masks the actual chunks in the text with their entity labels.
    parameters:
    ----------------
    faked: bool
    masked: bool
    merged_results: list of dict
    text: str
    fake_csv_path: str
    return:
    ----------------
    entities: list of dict
    masked_text: str
    faked_text: str
    """
    if faked and masked:
        text_mask = maskText(merged=merged_results, text=text)
        entities_with_faked_chunks = fakedChunk(fake_csv_path=fake_csv_path, merged=merged_results)
        faked_text = fakedText(merged=entities_with_faked_chunks, text=text)
        return {"entities": entities_with_faked_chunks, "masked_text":text_mask,"faked_text":faked_text}
    elif faked:
        entities_with_faked_chunks = fakedChunk(fake_csv_path=fake_csv_path,merged=merged_results)
        faked_text = fakedText(merged=entities_with_faked_chunks, text=text)
        return {"entities": entities_with_faked_chunks, "faked_text":faked_text}
    elif masked:
        text_mask = maskText(merged=merged_results, text=text)
        return {"entities": merged_results, "masked_text":text_mask}

    else:
        return {"entities": merged_results}
    

# TODO: Test this function
if __name__ == "__main__":
    pass


# Author: AIMPED
# Date: 2023-March-12
# Description: This file contains the regex parser for de-identification of clinical notes

def RegexNerParser(path, text, white_label_list):
    """
    Finds all the chunks that correspond to the regex pattern, 
    and checks their prefix and suffix collocations in the scope of context length.
    parameters:
    path: str
    text: str
    return:
    parser_results: list of dict
    """
    import json
    import re

    with open(path, 'r', encoding="utf8") as f:
        file = json.load(f)
    label = file["label"]
    parser_results = []
    if label in white_label_list:
        chunks = re.findall(file["regex"], text)

        for chunk in set(chunks):
            seek = 0
            for i in range(chunks.count(chunk)):
                strt = text.find(chunk, seek) - file["contextLength"]
                if strt < 0:
                    strt = 0
                before = text[strt:text.find(chunk, seek)]
                after = text[
                        text.find(chunk, seek) + len(chunk):text.find(chunk, seek) + len(chunk) + file["contextLength"]]
                begin = text.find(chunk, seek)
                end = text.find(chunk, seek) + len(chunk)
                seek = end
                left_context, right_context = [], []
                if file["prefix"]:
                    left_context = [True if pre.lower() in before.lower() else False for pre in file["prefix"]]

                if file["suffix"]:
                    right_context = [True if sfx.lower() in after.lower() else False for sfx in file["suffix"]]
                if any(left_context) or any(right_context):
                    parser_results.append(
                        {'chunk': chunk, 'confidence': 1, 'begin': begin, 'end': end, 'entity': file["label"]})
                elif (not bool(file["prefix"])) and (not bool(file["suffix"])):
                    parser_results.append(
                        {'chunk': chunk, 'confidence': 1, 'begin': begin, 'end': end, 'entity': file["label"]})
                else:
                    # parser_results.append({'chunk': chunk, 'confidence':0.67, 'begin': begin, 'end':end, 'entity':file["label"]})
                    left_context, right_context = [], []
                    continue

    return parser_results


def RegexModelNerMerger(rule, results_from_model):
    """
    Merges the results from regex and model.
    parameters:
    rule: list of dict
    results_from_model: list of dict
    return:
    merged: list of dict
    """
    import operator
    merged = []
    out = []
    if results_from_model and rule:
        out = [j for j in results_from_model for z in rule if
               j['begin'] <= z['begin'] < j['end'] or j['begin'] < z['end'] <= j['end'] or z['begin'] <= j['begin'] < z[
                   'end'] or z['begin'] < j['end'] <= z['end']]
        for removed in out:
            if removed in results_from_model:
                results_from_model.remove(removed)
        merged += results_from_model + rule
    else:
        merged += results_from_model + rule
    merged.sort(key=operator.itemgetter('begin'))
    return merged


def RegexModelNerChunkResults(regex_json_files_path_list, model_results, text, white_label_list):
    """Parses the text with regex and merges the results.
    parameters:
    ----------------
    regex_json_files_path_list: list of str
    model_results: list of dict
    text: str
    white_label_list: list of str
    return:
    ----------------
    merged_results: list of dict
    """
    merged = RegexNerParser(regex_json_files_path_list[0], text, white_label_list)
    for i in range(1, len(regex_json_files_path_list)):
        merged = RegexModelNerMerger(merged, RegexNerParser(regex_json_files_path_list[i], text, white_label_list))
    merged_results = RegexModelNerMerger(merged, model_results)
    return merged_results



# Author: AIMPED
# Date: 2023-March-11
# Description: NER model results

def NerModelResults(sents_tokens_list, sentences, tokenizer, model, text, device,
                    assertion_relation=False):
    """
    It returns the NER model results of a text.
    Parameters
    ----------
    sents_tokens_list : list
    tokenizer : transformers.PreTrainedTokenizer
    model : transformers.PreTrainedModel
    text : str
    device : torch.device
    assertion_relation : bool, optional
        The default is False.
    sentences : list, optional
        The default is []. Only used if assertion_relation is True

    Returns
    -------
    tokens : list
    preds : list
    probs : list
    begins : list
    ends : list
    sent_begins : list
        Only returned if assertion_relation is True
    sent_ends : list
        Only returned if assertion_relation is True
    sent_idxs : list
        Only returned if assertion_relation is True
    """
    import torch
    start = 0
    tokens, probs, begins, ends, preds, sent_begins, sent_ends, sent_idxs = [], [], [], [], [], [], [], []

    for sentence_idx, sent_token_list in enumerate(sents_tokens_list):
        start_sent = 0
        start = text.find(sentences[sentence_idx], start)
        model_inputs = tokenizer(sent_token_list, is_split_into_words=True, truncation=True,
                                 padding=False, max_length=512, return_tensors="pt").to(device)
        word_ids = model_inputs.word_ids()  # sub tokenlar sent_token_list deki hangi idxteki tokena ait
        # ornek word_ids = [None, 0, 1, 2, 2, 2, 3, 4, 5, 6, 7, 8, 9, 10, 10, 10, 11, 12, 13, 14, 15, None]
        outputs = model(**model_inputs)
        probabilities = torch.nn.functional.softmax(outputs.logits, dim=-1).tolist()[0]
        predictions = outputs.logits.argmax(dim=-1).tolist()[0]
        idx = 1
        while idx < len(word_ids) - 1:  # sondaki None icin islem yapmamak icin -1 yapildi
            word_id1 = word_ids[idx]
            word_id2 = word_ids[idx + 1]
            label = model.config.id2label[predictions[idx]]
            prob = max(probabilities[idx])
            if word_id1 == word_id2:
                while word_id1 == word_ids[idx]:
                    idx += 1
                idx -= 1

            token = sent_token_list[word_ids[idx]]
            begin = text.find(token, start)
            end = begin + len(token)
            tokens.append(token)
            begins.append(begin)
            ends.append(end)
            preds.append(label)
            probs.append(prob)
            start = end
            if assertion_relation:
                sentence_begin = sentences[sentence_idx].find(token, start_sent)
                sentence_end = sentence_begin + len(token)
                sent_begins.append(sentence_begin)
                sent_ends.append(sentence_end)
                sent_idxs.append(sentence_idx)
                start_sent = sentence_end
            idx += 1
    if not assertion_relation:
        return tokens, preds, probs, begins, ends
    else:
        return tokens, preds, probs, begins, ends, sent_begins, sent_ends, sent_idxs


# TODO : will be tested
if __name__ == "__main__":
    import torch
    import pandas as pd
    import numpy as np
    import os
    import re
    import json
    from tokenizer import sentence_tokenizer, word_tokenizer

    # load model
    from transformers import AutoTokenizer, AutoModelForTokenClassification

    saved_model_path = r"C:\Users\rcali\Desktop\kubeflow\deid-ner-gitlab\model"
    tokenizer = AutoTokenizer.from_pretrained(saved_model_path)
    model = AutoModelForTokenClassification.from_pretrained(saved_model_path)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    # load data
    text = open("aimped\\test\\data.txt", "r").read()
    sents_tokens_list = word_tokenizer(sentence_tokenizer(text, "english"))

    # ner = pipeline('token-classification', model=model, tokenizer=tokenizer, device=device)
    # print(ner(text))

    # print(sents_tokens_list)
    tokens, preds, probs, begins, ends = NerModelResults(sents_tokens_list, tokenizer, model, text, device)

    white_label_list = ['PATIENT', 'ORGANIZATION', 'SSN', 'SEX', 'DOCTOR', 'HOSPITAL', 'AGE', 'MEDICALRECORD', 'ZIP',
                        'STREET', 'EMAIL', 'DATE', 'ID', 'CITY', 'COUNTRY', 'PROFESSION']

    from chunk_merger import ChunkMerger

    merged_chunks = ChunkMerger(text, white_label_list, tokens, preds, probs, begins, ends)
    print(merged_chunks)



from transformers import (AutoModelForSequenceClassification,
                          AutoModelForTokenClassification, AutoTokenizer,
                          pipeline)
from aimped.nlp.pipeline import Pipeline
from aimped.nlp.tokenizer import sentence_tokenizer, word_tokenizer

# %%
ner_model_path = r"C:\Users\rcali\Desktop\kubeflow\relation\ner_model"
ner_tokenizer = AutoTokenizer.from_pretrained(ner_model_path)
ner_model = AutoModelForTokenClassification.from_pretrained(ner_model_path)

# %%
cls_model_path = r"C:\Users\rcali\Desktop\kubeflow\relation\cls_model"
cls_tokenizer = AutoTokenizer.from_pretrained(cls_model_path)
cs_model = AutoModelForSequenceClassification.from_pretrained(cls_model_path)

classifier = pipeline(task="sentiment-analysis", model=cs_model, tokenizer=cls_tokenizer, device=-1)
# %%
rel_model_path = r"C:\Users\rcali\Desktop\kubeflow\relation\rel_model"
rel_tokenizer = AutoTokenizer.from_pretrained(rel_model_path)
rel_model = AutoModelForSequenceClassification.from_pretrained(rel_model_path)

relation_classifier = pipeline(task="sentiment-analysis", model=rel_model, tokenizer=rel_tokenizer, device=-1)

# %%

text = '''Successful treatment with carbimazole	of a hyperthyroid pregnancy with  hepatic impairment  after propylthiouracil administration.
A 37-year-old female patient was diagnosed with hepatic impairment 8 months ago and medical treatment with carbimazole, low-dose corticosteroids and 5-ASA was started.
Successful treatment with carbimazole	of a hyperthyroid pregnancy with  hepatic impairment  after propylthiouracil administration.'''

pipe = Pipeline(model=ner_model, tokenizer=ner_tokenizer, device='cpu')
sentences = sentence_tokenizer(text, "english")
print(sentences)
sentences = [sentences[idx] for idx, label in enumerate(classifier(sentences)) if label['label'] == 'Positive']
sents_tokens_list = word_tokenizer(sentences)
print(sents_tokens_list)
white_label_list = ['DRUG', 'ADE']

tokens, preds, probs, begins, ends, sent_begins, sent_ends, sent_idxs = pipe.ner_result(text=text,
                                                                                        sents_tokens_list=sents_tokens_list,
                                                                                        sentences=sentences,
                                                                                        assertion_relation=True)
print("tokens: ", tokens)
print("preds: ", preds)
print("probs:", probs)
print("begins:", begins)
print("ends:", ends)
# %%
results = pipe.chunk_merger(text=text,
                            white_label_list=white_label_list, tokens=tokens, preds=preds, probs=probs, begins=begins,
                            ends=ends,
                            sent_begins=sent_begins, sent_ends=sent_ends, sent_idxs=sent_idxs, assertion_relation=True)
print("results: ", results)

# %%
relation_white_label_list = ['Positive', 'Negative']
results = pipe.relation_result(sentences=sentences, ner_chunk_results=results, relation_classifier=relation_classifier,
                               ner_white_label_list=white_label_list,
                               relation_white_label_list=relation_white_label_list,
                               one_to_many=True, one_label='DRUG', return_svg=False)
print("results: ", results)



# Author: AIMPED
# Date: 2023-March-11
# Description: This file contains model loading functions

def load_config(file_path='config.json'):
    import json
    with open(file_path, "r") as f:
        config = json.load(f)
    return config

def load_model(model_name):
    from transformers import AutoTokenizer, AutoModelForTokenClassification
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model = AutoModelForTokenClassification.from_pretrained(model_name)
    return tokenizer, model


# Author: AIMPED
# Date: 2023-March-11
# Description: This file contains utility functions

import json
import logging
import numpy as np
from .version import __version__


def get_version():
    """Returns the version of aimped library."""
    return f'aimped version: {__version__}'

def get_handler(log_file='KSERVE.log', log_level=logging.DEBUG):
    """Returns the logger object for logging the output of the server.
    parameters:
    ----------------
    log_file: str
    log_level: logging level
    return:
    ----------------
    logger: logging object
    """

    f_handler = logging.FileHandler(log_file, encoding='utf-8', mode='w')
    formatter = logging.Formatter(
        '[%(asctime)s %(filename)s:%(lineno)s] - %(message)s')
    f_handler.setFormatter(formatter)
    f_handler.setLevel(log_level)

    logger = logging.getLogger()
    logger.setLevel(log_level)
    logger.addHandler(f_handler)

    return logger

def cuda_info():
    """Returns the cuda information if cuda is available."""
    import torch
    try:
        if torch.cuda.is_available():
            cuda_info = {"cuda is available": torch.cuda.is_available(),
                         "device count": torch.cuda.device_count(),
                         "current device": torch.cuda.current_device(),
                         "device name": torch.cuda.get_device_name(0),
                         "Memory Usage": {"Allocated": round(torch.cuda.memory_allocated(0)/1024**3,1),
                                          "Cached": round(torch.cuda.memory_reserved(0)/1024**3,1)}}
            return cuda_info
        else:
            return "cuda not available"
    except:
        return "cuda information not available"


class NumpyFloatValuesEncoder(json.JSONEncoder):
    """This class is used to convert numpy float32 to float"""
    def default(self, obj):
        if isinstance(obj, np.float32):
            return float(obj)
        return json.JSONEncoder.default(self, obj)


# TODO: test these functions
if __name__ == '__main__':
    print(get_version())


from aimped.nlp.tokenizer import sentence_tokenizer, word_tokenizer


def test_sentence_tokenizer():
    text = "This is a test. This is a test."
    sents = sentence_tokenizer(text, language="english")
    print(sents)
    assert sents == ['This is a test.', 'This is a test.']

def test_word_tokenizer():
    text = "This is a test. This is a test."
    words = word_tokenizer(sentence_tokenizer(text, language="english"))
    print(words)
    assert words == [['This', 'is', 'a', 'test', '.'], ['This', 'is', 'a', 'test', '.']]

if __name__ == "__main__":
    test_sentence_tokenizer()
    test_word_tokenizer()



import torch
from aimped.nlp.tokenizer import sentence_tokenizer, word_tokenizer
from transformers import AutoTokenizer, AutoModelForTokenClassification
from aimped.nlp.pipeline import Pipeline

# load model
saved_model_path = r"C:\Users\rcali\Desktop\kubeflow\deid-ner-gitlab\model"
tokenizer = AutoTokenizer.from_pretrained(saved_model_path)
model = AutoModelForTokenClassification.from_pretrained(saved_model_path)
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
# print(f"device: {device}")

# load data
text = open("data.txt", "r").read()
sentences = sentence_tokenizer(text, "english")
sents_tokens_list = word_tokenizer(sentences)
white_label_list = ['PATIENT', 'ORGANIZATION', 'SSN', 'SEX', 'DOCTOR', 'HOSPITAL', 'AGE', 'MEDICALRECORD', 'ZIP',
                    'STREET', 'EMAIL', 'DATE', 'ID', 'CITY', 'COUNTRY', 'PROFESSION']
# print(sents_tokens_list)
pipe = Pipeline(model=model, tokenizer=tokenizer, device='cpu')
tokens, preds, probs, begins, ends = pipe.ner_result(text=text,
                                                     sents_tokens_list=sents_tokens_list,
                                                     sentences=sentences)

merged_chunks = pipe.chunk_merger(text, white_label_list, tokens, preds, probs, begins, ends)
# print("merged_chunks: ", merged_chunks)

regex_json_files_path = r"C:\Users\rcali\Desktop\kubeflow\deid-ner-gitlab\nlp-health-deidentification-sub-base-en\json_regex"
merged_results = pipe.regex(regex_json_files_path, merged_chunks, text, white_label_list)
print("merged_results:", merged_results)



# Author AIMPED
# Date 2023-March-12
# Description This file contains the pipeline for assertion detection of clinical notes


def assertion_annotate_sentence(df):
    """
    It annotates the sentence with [Entity] tags.   
    parameters:
    ----------------
    df: pandas dataframe
    return:
    ----------------
    df['new_sentence']: str
    """
    import pandas as pd

    df['new_sentence'] = " ".join([
        df['sentence'][:df['sent_begin']],
        ' [Entity] ',
        df['sentence'][df['sent_begin']:df['sent_end']],
        ' [Entity] ',
        df['sentence'][df['sent_end']:]])
    return df['new_sentence']


def AssertionModelResults(ner_results, sentences, classifier, assertion_white_label_list):
    """
    It returns the assertion detection results of a text.
    parameters:
    ----------------
    ner_results: list of dict
    sentences: list of str
    tokenizer: transformers.tokenization_utils_base.PreTrainedTokenizer
    model: transformers.modeling_utils.PreTrainedModel
    classifier: transformers.modeling_utils.PreTrainedModel
    return:
    ----------------
    df.to_dict(orient = 'records'): list of dict
    """
    import pandas as pd
    import numpy as np
    df = pd.DataFrame()
    if len(ner_results) >= 1:
        ner_results = list(map(lambda x: list(x.values()), ner_results))
        df = pd.DataFrame(ner_results)
        df.columns = ['ner_label', 'chunk', 'begin', 'end', 'sent_idx', 'sent_begin', 'sent_end']
        if len(df) != 0:
            df['sentence'] = np.nan
            for i in df.sent_idx.unique():
                df.loc[df[df.sent_idx == i].index, 'sentence'] = sentences[i]
            df['new_sentence'] = np.nan
            df['new_sentence'] = df.apply(assertion_annotate_sentence, axis=1)
            df.reset_index(drop=True, inplace=True)
            rel_results = classifier(list(df['new_sentence']))
            df = pd.concat([df, pd.DataFrame(rel_results)], axis=1)
            df = df[['begin', 'end', 'ner_label', 'chunk', 'label', 'score']]
            df.columns = ['begin', 'end', 'ner_label', 'chunk', 'assertion', 'score']
            df = df[df['assertion'].isin(assertion_white_label_list)]
    return df.to_dict(orient='records')



# Author: AIMPED
# Date: 2023-March-10
# Description: Text tokenizer


def sentence_tokenizer(text:str, language:str)->list:
    """ 
    Tokenize a text into sentences.
    text: str    
    language: str (turkish, swedish, spanish, slovene, 
            portuguese, polish, norwegian, italian, 
            greek, german, french, finnish, estonian 
            english, dutch, danish, czech )
    """    
    import nltk
    sent = nltk.tokenize.sent_tokenize(text, language=language)
    return sent

def word_tokenizer(sentences:list)->list:
    """
    Tokenize a list of sentences into words.
    parameters:
    sentences: list of str
    return:
    tokens: list of list of str
    """
    import nltk                      
    tokens= [nltk.tokenize.wordpunct_tokenize(i) for i in sentences]
    return tokens


# TODO: Test these functions
if __name__ == "__main__":
    text = "This is a test. This is only a test."
    print(sentence_tokenizer(text, "english"))
    print(word_tokenizer(sentence_tokenizer(text, "english")))


from transformers import (AutoModelForSequenceClassification,
                          AutoModelForTokenClassification, AutoTokenizer,
                          pipeline)
from aimped.nlp.pipeline import Pipeline
from aimped.nlp.tokenizer import sentence_tokenizer, word_tokenizer

text = '''Patient has a headache for the last 2 weeks, needs to get a head CT, and appears anxious when she walks fast. 
No alopecia noted. She denies pain.
Patient has a headache for the last 2 weeks, needs to get a head CT, and appears anxious when she walks fast. 
No alopecia noted. She denies pain.'''

as_model_path = r"C:\Users\rcali\Desktop\kubeflow\ner_assertion\assertion_model"
as_tokenizer = AutoTokenizer.from_pretrained(as_model_path)
as_model = AutoModelForSequenceClassification.from_pretrained(as_model_path)
classifier = pipeline(task="sentiment-analysis", model=as_model, tokenizer=as_tokenizer, device=-1)

ner_model_path = r"C:\Users\rcali\Desktop\kubeflow\ner_assertion\ner_model"
ner_tokenizer = AutoTokenizer.from_pretrained(ner_model_path)
ner_model = AutoModelForTokenClassification.from_pretrained(ner_model_path)

ner_pipe = Pipeline(model=ner_model, tokenizer=ner_tokenizer, device='cpu')

sentences = sentence_tokenizer(text, "english")
sents_tokens_list = word_tokenizer(sentences)
white_label_list = ['problem', 'test', 'treatment']
assertion_white_label_list = ['present', 'absent', 'possible']

# print("sentences: ", sentences)
# print("sents_tokens_list: ", sents_tokens_list)

tokens, preds, probs, begins, ends, sent_begins, sent_ends, sent_idxs = ner_pipe.ner_result(
    text=text,
    sents_tokens_list=sents_tokens_list,
    sentences=sentences,
    assertion_relation=True
)
# print(tokens, preds, probs, begins, ends, sent_begins, sent_ends, sent_idxs)

ner_results = ner_pipe.chunk_merger(text=text,
                                    white_label_list=white_label_list,
                                    tokens=tokens,
                                    preds=preds,
                                    probs=probs,
                                    begins=begins,
                                    ends=ends,
                                    assertion_relation=True,
                                    sent_begins=sent_begins,
                                    sent_ends=sent_ends,
                                    sent_idxs=sent_idxs)

# print(ner_results)

results = ner_pipe.assertion_result(ner_results=ner_results,
                                    classifier=classifier,
                                    assertion_white_label_list=assertion_white_label_list,
                                    sentences=sentences,
                                    )

print(results)



# Author: AIMPED    
# Date: 2023-March-10
# Description: Text cleaner


def text_cleaner(text:str, 
                 language:str='english', 
                 remove_stopwords:bool=True, 
                 remove_short_words:int=1, 
                 remove_numbers:bool=True, 
                 remove_punctuation:bool=True, 
                 remove_non_ascii:bool=True, 
                 lower_case:bool=True)->str:
    
    """Clean text by removing unnecessary characters and altering the format of words.
    :param text: The text to be cleaned
    :param language: The language of the text
    :param remove_stopwords: Whether to remove stopwords or not
    :param remove_short_words: Whether to remove short words or not
    :param remove_numbers: Whether to remove numbers or not
    :param remove_punctuation: Whether to remove punctuation or not
    :param remove_non_ascii: Whether to remove non-ascii characters or not
    :param lower_case: Whether to convert all characters to lower case or not
    :return: The cleaned text
    """

    import re
    import nltk

    # lower case
    if lower_case: 
        text = text.lower()
    # remove punctuation
    if remove_punctuation:
        text = re.sub(r'[^\w\s]','',text)
    # remove numbers
    if remove_numbers:
        text = re.sub(r'\d+','',text)
    # remove stop words
    if remove_stopwords:
        try:
            stopwords = nltk.corpus.stopwords.words(language)
            text = ' '.join([word for word in text.split() if word not in stopwords])
        except LookupError:
            nltk.download('stopwords')
            stopwords = nltk.corpus.stopwords.words(language)
            text = ' '.join([word for word in text.split() if word not in stopwords])
        except:
            print('No stopwords for language: {}'.format(language))
            print("Please use nltk.download('stopwords') to download the stopwords for the language you want to use.")
    # remove short words
    if remove_short_words:
        text = ' '.join([word for word in text.split() if len(word) > remove_short_words])
    # remove non-ascii characters
    if remove_non_ascii:
        text = text.encode('ascii', 'ignore').decode('utf-8', 'ignore')
    return text

if __name__ == "__main__":
    text = "This is an easy test°′﻿, and This is the last chance."
    print(text_cleaner(text))


from aimped import nlp
from aimped import version
from aimped import utils
from aimped import nitro
from aimped import dataset
from aimped import models
from aimped import train
from aimped import test
from aimped.version import __version__



from aimped.nlp import assertion
from aimped.nlp import chunk_merger
from aimped.nlp import cleaner
from aimped.nlp import deid
from aimped.nlp import ner
from aimped.nlp import regex_parser
from aimped.nlp import tokenizer
from aimped.nlp import relation
from aimped.nlp import pipeline

# from aimped.nlp.pipeline import *
# from aimped.nlp.assertion import *
# from aimped.nlp.chunk_merger import *
# from aimped.nlp.cleaner import *
# from aimped.nlp.deid import *
# from aimped.nlp.ner import *
# from aimped.nlp.regex_parser import *
# from aimped.nlp.tokenizer import *
# from aimped.nlp.relation import *



import setuptools
from aimped.version import __version__

setuptools.setup(
    name="aimped",
    version= __version__,
    packages=setuptools.find_packages(),
    install_requires=[ 
                        'nltk',
                        'numpy',
                        'pandas',
                        'scikit-learn',
                        "seqeval"
                        ],
    author="Russell C., Raife H.",
    author_email="russell@aimped.com, raife@aimped.com",
    maintainer="aimped",
    maintainer_email="contact@aimped.com",
    description="Aimped is a unique library that provides classes and functions for only exclusively business-tailored AI-based NLP models.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://dev.ml-hub.nioyatechai.com/", 
    
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)




# Author AIMPED
# Date 2023-March-11
# Description: This file contains the pipeline wrapper for NER, Assertion and De-identification models.


class Pipeline:
    """
    It returns the ner results of a text.
    parameters:
    ----------------
    task: str
    text: str
    model: transformers.modeling_utils.PreTrainedModel
    tokenizer: transformers.tokenization_utils_base.PreTrainedTokenizer
    *args: arguments
    **kwargs: keyword arguments
    return:
    ----------------
    results: list of dict
    """

    def __init__(self, tokenizer, model, device='cpu'):
        """Initialize the pipeline class."""
        self.tokenizer = tokenizer
        self.model = model
        self.device = device

    def ner_result(self, text, sents_tokens_list, sentences, assertion_relation=False):
        """It returns the ner results of a text.
        parameters:
        ----------------
        text: str
        sents_tokens_list: list of list of str
        assertion_relation: bool
        sentences: list of str
        return:
        ----------------
        ner_results: list of dict"""
        from aimped.nlp.ner import NerModelResults
        ner_results = NerModelResults(text=text,
                                      tokenizer=self.tokenizer,
                                      model=self.model,
                                      device=self.device,
                                      sents_tokens_list=sents_tokens_list,
                                      sentences=sentences,
                                      assertion_relation=assertion_relation
                                      )

        return ner_results

    def deid_result(self, text, merged_results, fake_csv_path, faked=False, masked=False):
        """It returns the deid results of a text.
        parameters:
        ----------------
        text: str
        merged_results: list of dict
        fake_csv_path: str
        faked: bool
        masked: bool
        return:
        ----------------
        results: list of dict
        """
        from aimped.nlp.deid import maskText, fakedChunk, fakedText, deidentification
        results = deidentification(text=text,
                                   merged_results=merged_results,
                                   fake_csv_path=fake_csv_path,
                                   faked=faked,
                                   masked=masked)
        return results

    def assertion_result(self, ner_results, sentences, classifier, assertion_white_label_list):
        """It returns the assertion results of a text.
        parameters:
        ----------------
        text: str
        ner_results: list of dict
        sentences: list of str
        classifier: str
        return:
        ----------------
        results: list of dict
        """
        from aimped.nlp.assertion import assertion_annotate_sentence, AssertionModelResults
        results = AssertionModelResults(ner_results=ner_results,
                                        sentences=sentences,
                                        classifier=classifier,
                                        assertion_white_label_list=assertion_white_label_list
                                        )
        return results

    def chunk_merger(self, text, white_label_list, tokens, preds, probs, begins, ends,
                     assertion_relation=False, sent_begins=[], sent_ends=[], sent_idxs=[]):
        """It returns the merged chunks of a text.
        parameters:
        ----------------
        text: str
        white_label_list: list of str
        tokens: list of str
        preds: list of str
        probs: list of float
        begins: list of int
        ends: list of int
        assertion_relation: bool
        sent_begins: list of int
        sent_ends: list of int
        sent_idxs: list of int
        return:
        ----------------
        results: list of dict
        """
        from aimped.nlp.chunk_merger import ChunkMerger
        results = ChunkMerger(text=text,
                              white_label_list=white_label_list,
                              tokens=tokens,
                              preds=preds,
                              probs=probs,
                              begins=begins,
                              ends=ends,
                              assertion_relation=assertion_relation,
                              sent_begins=sent_begins,
                              sent_ends=sent_ends,
                              sent_idxs=sent_idxs)
        return results

    def regex(self, regex_json_files_path, model_results, text, white_label_list):
        """It returns the regex results of a text.
        parameters:
        ----------------
        regex_json_files_path: str
        model_results: list of dict
        text: str
        white_label_list: list of str
        return:
        ----------------
        results: list of dict
        """
        from aimped.nlp.regex_parser import RegexNerParser, RegexModelNerMerger, RegexModelNerChunkResults
        import glob
        regex_json_files_path_list = glob.glob(f"{regex_json_files_path}/*.json")

        merged_results = RegexModelNerChunkResults(regex_json_files_path_list=regex_json_files_path_list,
                                                   model_results=model_results,
                                                   text=text,
                                                   white_label_list=white_label_list)
        return merged_results

    def relation_result(self, sentences, ner_chunk_results, relation_classifier,
                        ner_white_label_list, relation_white_label_list, one_to_many=True,
                        one_label=None, return_svg=False):
        """It returns the relation results of a text.
        parameters:
        ----------------
        sentences: list of str
        ner_chunk_results: list of dict
        relation_classifier: str
        ner_white_label_list: list of str
        relation_white_label_list: list of str
        one_to_many: bool = True
        one_label: str = None
        return_svg: bool = False
        return:
        ----------------
        results: list of dict
        """
        from aimped.nlp.relation import RelationResults, relation_annotate_sentence
        results = RelationResults(sentences=sentences,
                                  ner_chunk_results=ner_chunk_results,
                                  relation_classifier=relation_classifier,
                                  ner_white_label_list=ner_white_label_list,
                                  relation_white_label_list=relation_white_label_list,
                                  one_to_many=one_to_many,
                                  one_label=one_label,
                                  return_svg=return_svg)
        return results

    def __str__(self) -> str:
        """Return the string representation of the pipeline."""
        return f"Pipeline(model={self.model}, tokenizer={self.tokenizer})"


# TEST CODE
if __name__ == "__main__":
    import torch
    import pandas as pd
    import numpy as np
    import os
    import re
    import json
    from tokenizer import sentence_tokenizer, word_tokenizer

    # load model
    from transformers import AutoTokenizer, AutoModelForTokenClassification

    saved_model_path = r"C:\Users\rcali\Desktop\kubeflow\deid-ner-gitlab\model"
    tokenizer = AutoTokenizer.from_pretrained(saved_model_path)
    model = AutoModelForTokenClassification.from_pretrained(saved_model_path)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    # print(f"device: {device}")

    # load data
    text = open("aimped\\test\\data.txt", "r").read()
    sentences = sentence_tokenizer(text, "english")
    sents_tokens_list = word_tokenizer(sentences)
    white_label_list = ['PATIENT', 'ORGANIZATION', 'SSN', 'SEX', 'DOCTOR', 'HOSPITAL', 'AGE', 'MEDICALRECORD', 'ZIP',
                        'STREET', 'EMAIL', 'DATE', 'ID', 'CITY', 'COUNTRY', 'PROFESSION']
    # print(sents_tokens_list)

    # test pipeline
    pipe = Pipeline(model=model, tokenizer=tokenizer, device='cpu')
    tokens, preds, probs, begins, ends = pipe.ner_result(text=text,
                                                         sents_tokens_list=sents_tokens_list,
                                                         sentences=sentences)
    # print("tokens: ", tokens)
    # print("preds: ", preds)
    # print("probs:", probs)
    # print("begins:", begins)
    # print("ends:", ends)
    merged_chunks = pipe.chunk_merger(text, white_label_list, tokens, preds, probs, begins, ends)
    # print(merged_chunks)

    # test regex
    regex_json_files_path = r"C:\Users\rcali\Desktop\kubeflow\deid-ner-gitlab\nlp-health-deidentification-sub-base-en\json_regex"
    merged_results = pipe.regex(regex_json_files_path, merged_chunks, text, white_label_list)
    # print(merged_results)

    # test deid
    fake_csv_path = r"C:\Users\rcali\Desktop\kubeflow\deid-ner-gitlab\nlp-health-deidentification-sub-base-en\fake.csv"
    deid_results = pipe.deid_result(text, merged_results, fake_csv_path, faked=True, masked=True)
    import pprint

    pprint.pprint(deid_results, indent=1)

    # test assertion



# Author: AIMPED
# Date: 2023-March-11
# Description: Chunk merger


def ChunkMerger(text, white_label_list, tokens, preds, probs, begins, ends,
                assertion_relation=False, sent_begins=[], sent_ends=[], sent_idxs=[]):
    """
    Parameters
    ----------
    text : str
    white_label_list : list
    tokens : list
    preds : list
    probs : list
    begins : list
    ends : list
    assertion_relation : bool, optional
        The default is False.
    sent_begins : list, optional
        The default is []. Only used if assertion_relation is True
    sent_ends : list, optional
        The default is []. Only used if assertion_relation is True
    sent_idxs : list, optional
        The default is []. Only used if assertion_relation is True

    Returns
    -------
    results : list
    """
    import numpy as np
    results = []
    idx = 0
    while idx < len(tokens):
        label = preds[idx]
        if label != "O": label = label[2:]
        if label in white_label_list:
            all_scores = []
            if preds[idx][2:] == label and preds[idx - 1 if idx != 0 else idx][2:] != label or idx == 0:
                all_scores.append(probs[idx])
                if assertion_relation:
                    begin, end, sent_begin, sent_end = begins[idx], ends[idx], sent_begins[idx], sent_ends[idx]
                else:
                    begin, end = begins[idx], ends[idx]
            if preds[idx][2:] == label and preds[idx - 1 if idx != 0 else idx][2:] == label and idx != 0:
                while idx < len(tokens) and preds[idx][2:] == label:
                    all_scores.append(probs[idx])
                    if assertion_relation:
                        end, sent_end = ends[idx], sent_ends[idx]
                    else:
                        end = ends[idx]
                    idx += 1
                idx -= 1

            if idx == len(tokens) - 1:
                if not assertion_relation:
                    confidence = np.mean(all_scores).item()
                    chunk = text[begin:end]
                    results.append({"entity": label,
                                    "confidence": confidence,
                                    "chunk": chunk,
                                    "begin": int(begin),
                                    "end": int(end)})
                else:
                    chunk = text[begin:end]
                    results.append({"entity": label,
                                    "chunk": chunk,
                                    "begin": int(begin),
                                    "end": int(end),
                                    "sent_idx": int(sent_idxs[idx]),
                                    "sent_begin": int(sent_begin),
                                    "sent_end": int(sent_end)})


            elif idx < len(tokens) - 1:
                if preds[idx + 1][2:] != label:
                    if not assertion_relation:
                        confidence = np.mean(all_scores).item()
                        chunk = text[begin:end]
                        results.append({"entity": label,
                                        "confidence": confidence,
                                        "chunk": chunk,
                                        "begin": int(begin),
                                        "end": int(end)})
                    else:
                        chunk = text[begin:end]
                        results.append({"entity": label,
                                        "chunk": chunk,
                                        "begin": int(begin),
                                        "end": int(end),
                                        "sent_idx": int(sent_idxs[idx]),
                                        "sent_begin": int(sent_begin),
                                        "sent_end": int(sent_end)})

        idx += 1
    return results


if __name__ == "__main__":
    pass
    # results = ChunkMerger()
    # print(results)



# Author AIMPED
# Date 2023-March-14
# Description This file contains the pipeline for relation extraction.

def relation_annotate_sentence(df):
    """It annotates the sentence with [Entity] tags.
    parameters:
    ----------------
    df: pandas dataframe
    return:
    ----------------
    df['new_sentence']: str
    """

    a1_start = min(df['sent_begin1'] - 1, df['sent_begin2'])
    a1_end = min(df['sent_end1'] + 1, df['sent_end2'] + 1)

    a2_start = max(df['sent_begin1'] - 1, df['sent_begin2'])
    a2_end = max(df['sent_end1'] + 1, df['sent_end2'] + 1)

    df['new_sentence'] = " ".join([
        df['sentence'][:a1_start],
        'e1b',
        df['sentence'][a1_start:a1_end],
        'e1e',
        df['sentence'][a1_end:a2_start],
        'e2b',
        df['sentence'][a2_start:a2_end],
        'e2e',
        df['sentence'][a2_end:]
    ])

    return df['new_sentence']


def RelationResults(sentences, ner_chunk_results, relation_classifier,
                    ner_white_label_list, relation_white_label_list, one_to_many,
                    one_label, return_svg):
    """It returns the relation results of a text.
    parameters:
    ----------------
    sentences: list of str
    ner_chunk_results: list of dict
    relation_classifier: str
    ner_white_label_list: list of str
    relation_white_label_list: list of str
    one_to_many: bool
    one_label: str
    return_svg: bool
    return:
    ----------------
    results: list of dict
    """
    import itertools
    import pandas as pd
    import numpy as np
    end_df = pd.DataFrame()
    df_ner_chunk_results = pd.DataFrame(ner_chunk_results)
    for i in df_ner_chunk_results.sent_idx.unique():
        sentence_bazli_ner_results = df_ner_chunk_results[df_ner_chunk_results.sent_idx == i]
        sentence_bazli_ner_results = sentence_bazli_ner_results.values.tolist()
        if len(sentence_bazli_ner_results) >= 2:
            df = pd.DataFrame(itertools.combinations(sentence_bazli_ner_results, 2))
            df['firstCharEnt1'] = df[0].apply(lambda x: x[2])
            df['lastCharEnt1'] = df[0].apply(lambda x: x[3])
            df['entity1'] = df[0].apply(lambda x: x[0])
            df['chunk1'] = df[0].apply(lambda x: x[1])
            df['sent_begin1'] = df[0].apply(lambda x: x[-2])
            df['sent_end1'] = df[0].apply(lambda x: x[-1])
            df['firstCharEnt2'] = df[1].apply(lambda x: x[2])
            df['lastCharEnt2'] = df[1].apply(lambda x: x[3])
            df['entity2'] = df[1].apply(lambda x: x[0])
            df['chunk2'] = df[1].apply(lambda x: x[1])
            df['sent_begin2'] = df[1].apply(lambda x: x[-2])
            df['sent_end2'] = df[1].apply(lambda x: x[-1])

            if one_to_many:
                df = df[((df.entity1 == one_label) | (df.entity2 == one_label))]
                df = df[~((df.entity1 == one_label) & (df.entity2 == one_label))]
            else:
                for ner_label in ner_white_label_list:
                    df = df[~((df.entity1 == ner_label) & (df.entity2 == ner_label))]

            if len(df) != 0:
                df['sentID'] = i
                df['sentence'] = sentences[i]
                df = df.drop([0, 1], axis=1)
                df['new_sentence'] = np.nan
                df['new_sentence'] = df.apply(relation_annotate_sentence, axis=1)
                df.reset_index(drop=True, inplace=True)
                rel_results = relation_classifier(list(df['new_sentence']))
                df = pd.concat([df, pd.DataFrame(rel_results)], axis=1)
                df = df[['sentID', 'sentence', 'firstCharEnt1', 'sent_begin1', 'lastCharEnt1', 'sent_end1', 'entity1',
                         'chunk1',
                         'firstCharEnt2', 'sent_begin2', 'lastCharEnt2', 'sent_end2', 'entity2', 'chunk2', 'label',
                         'score'
                         ]]
                df = df[df['label'].isin(relation_white_label_list)]
                end_df = pd.concat([end_df, df], ignore_index=True)

    if end_df.empty or return_svg:
        return end_df.to_dict(orient='records')

    else:
        return end_df[['firstCharEnt1', 'lastCharEnt1', 'entity1', 'chunk1',
                       'firstCharEnt2', 'lastCharEnt2', 'entity2', 'chunk2', 'label', 'score']].to_dict(
            orient='records')
